@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico
-   Copyright (C) 2007-2008, 2010, 2012, 2016 Sergey Poznyakoff
+   Copyright (C) 2007-2008, 2010, 2012, 2016-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1749,8 +1749,13 @@ urf8_symcasecmp(char *a, char *b)
     return 0;
 }
 
+enum {
+    case_sensitive,
+    case_insensitive
+};
+
 int
-utf8_strcasecmp(char *a, char *b)
+utf8_strcmp_cc(char *a, char *b, int ci)
 {
     int alen, blen;
 
@@ -1768,8 +1773,10 @@ utf8_strcasecmp(char *a, char *b)
 	if (blen == 0)
 	    return 1;
 	utf8_mbtowc(&wb, b, blen);
-	wa = utf8_wc_toupper(wa);
-	wb = utf8_wc_toupper(wb);
+	if (ci == case_insensitive) {
+	    wa = utf8_wc_toupper(wa);
+	    wb = utf8_wc_toupper(wb);
+	}
 	if (wa < wb)
 	    return -1;
 	if (wa > wb)
@@ -1781,6 +1788,18 @@ utf8_strcasecmp(char *a, char *b)
     return 0;
 }
 
+int
+utf8_strcmp(char *a, char *b)
+{
+    return utf8_strcmp_cc(a, b, case_sensitive);
+}
+
+int
+utf8_strcasecmp(char *a, char *b)
+{
+    return utf8_strcmp_cc(a, b, case_insensitive);
+}
+
 int
 utf8_strncasecmp(char *a, char *b, size_t maxlen)
 {
@@ -1818,7 +1837,70 @@ utf8_strncasecmp(char *a, char *b, size_t maxlen)
     return 0;
 }
 
+#define is_alnumspace(c) (utf8_wc_is_alnum(c) || utf8_wc_is_space(c))
+int
+utf8_strcmp_alnumspace_cc(char *a, char *b, int ci)
+{
+    int alen, blen;
+    unsigned wa, wb;
 
+    while (*a) {	
+	alen = utf8_char_width(a);
+	if (alen == 0)
+	    return -1;
+	utf8_mbtowc(&wa, a, alen);
+	a += alen;
+
+	if (is_alnumspace(wa)) {
+	    if (*b == 0)
+		return 1;
+	    while (*b) {
+		blen = utf8_char_width(b);
+		if (blen == 0)
+		    return 1;
+		utf8_mbtowc(&wb, b, blen);
+		b += blen;
+		
+		if (is_alnumspace(wb)) {
+		    if (ci == case_insensitive) {
+			wa = utf8_wc_toupper(wa);
+			wb = utf8_wc_toupper(wb);
+		    }
+		    if (wa < wb)
+			return -1;
+		    if (wa > wb)
+			return 1;
+		    break;
+		}
+	    }
+	}
+    }
+
+    while (*b) {
+	blen = utf8_char_width(b);
+	if (blen == 0)
+	    return 1;
+	utf8_mbtowc(&wb, b, blen);
+	b += blen;
+	if (is_alnumspace(wb))
+	    return -1;
+    }
+    
+    return 0;
+}
+
+int
+utf8_strcmp_alnumspace(char *a, char *b)
+{
+    return utf8_strcmp_alnumspace_cc(a, b, case_sensitive);
+}
+
+int
+utf8_strcasecmp_alnumspace(char *a, char *b)
+{
+    return utf8_strcmp_alnumspace_cc(a, b, case_insensitive);
+}
+
 unsigned
 utf8_wc_toupper(unsigned wc)
 {
@@ -2274,3 +2356,29 @@ utf8_quote(const char *str, char **sptr)
     }
     return rc;
 }
+
+int
+utf8_table_check(unsigned wc,
+		 unsigned const *start, int const *count, size_t size)
+{
+    const unsigned *l = start;
+    const unsigned *r = start + size;
+    const unsigned *s = NULL;
+
+    if (wc == 0)
+	return 0;
+
+    while (l < r) {
+	s = l + ((r - l) >> 1);
+
+	if (*s <= wc) {
+	    l = s + 1;
+	} else {
+	    r = s;
+	}
+    }
+    
+    --l;
+    
+    return (wc < l[0] + count[l - start]);
+}
